/*
 * Telemetry.cpp: part of PQTelemetry v5 for VxWorks. 
 * 
 * Copyright (c) June 2013, Daniel Richman
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * To facilitate integrating this software into your own projects, redistributions in binary form need not retain any of this
 * 		license text, but the below disclaimer still applies. Attribution would be appreciated nonetheless!
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "Telemetry.h"

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "sockLib.h"

const unsigned short kTelemUDPDestPort = 17171; // PenguinGrapher receives on this port

const size_t kTelemPacketDefaultSize = 64 * 32; // # of bytes in starting packet size
const size_t kTelemPacketExpansionIncrement = 8 * 32; // increment to expand packet size by, if needed
const size_t kTelemPacketMaxSize = 192 * 32; // to prevent memory overflows, Telem packets are limited to this size

Telemetry* Telemetry::mInstance = NULL;

Telemetry* Telemetry::getSingletonPtr() {

	if (mInstance == NULL) {
		mInstance = new Telemetry();
	}

	return mInstance;
}

void Telemetry::init(char *serverIP, Timer *timer) {

#ifndef ENABLE_TELEMETRY
	return;
#endif

	// Were we passed in a timer, or should we make our own?
	if (timer != NULL) {
		mTimer = timer;
	} else {
		mTimer = new Timer();
		mTimer->Start();
	}

	/* set up network stuff: create a UDP socket */
	if ((sFd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == ERROR) {
		printf("construct socket error!\n");
	}

	/* explicit bind not required - local port number is dynamic */

	/* build server socket address */

	sockAddrSize = sizeof(struct sockaddr_in);
	bzero((char *) &serverAddr, sockAddrSize);
	serverAddr.sin_len = (u_char)sockAddrSize;
	serverAddr.sin_family = AF_INET;
	serverAddr.sin_port = htons(kTelemUDPDestPort);
	if (inet_aton(serverIP, &serverAddr.sin_addr) == ERROR) {
		printf("Error in Telemetry.\n");
		close(sFd);
	}

	/*
	 * Set up a mutual exclusion semaphore (mutex), which allows us to have nested semTakes if they're in the same thread. 
	 * startNewPacket(), which requires the semaphore, is called by either init() or sendPacket(). 
	 * 
	 * SEM_INVERSION_SAFE makes sure a high-priority caller of Telemetry isn't kept waiting. See VxWorks semaphore guide for more. 
	 */
	mPacketSem = semMCreate(SEM_Q_PRIORITY | SEM_INVERSION_SAFE | SEM_DELETE_SAFE);

	mBytesUsed = 0;
	mAllocatedPacketSize = kTelemPacketDefaultSize;
	startNewPacket(); // set up the first packet

	semGive(mPacketSem); // release the semaphore
}

void Telemetry::startNewPacket() {
#ifndef ENABLE_TELEMETRY
	return;
#endif

	semTake(mPacketSem, WAIT_FOREVER);

	/* The first time through, mAllocatedPacketSize will have been
	 * set to kDefaultMaxTelemVariables by the constructor. But
	 * if we needed more space at some point, start with that size now. */
	mVarsStart = (char *) malloc(mAllocatedPacketSize); // double takes 8 bytes

	// Reset mVarCount: have to do this BEFORE addVariable so it doesn't get hopelessly confused
	mBytesUsed = 0;

	// Copy the time as the first variable
	// The special name "time" is used internally by PQTelemetry and should not be used for any other variable. 
	addVariable("time", mTimer->Get());

	semGive(mPacketSem);
}

void Telemetry::addVariable(char *variableName, double value) {

#ifndef ENABLE_TELEMETRY
	return;
#endif

	// Make sure no one tries to modify the array while add the variable
	semTake(mPacketSem, WAIT_FOREVER);

	/* Make sure we have room in memory. 
	 * If insufficient room (actually, we check early), expand the packet. 
	 * 
	 * Note: size check is (current size used + variable name length + 1 byte for the ':' required by PQTelemetry
	 * + 8 bytes for the double value. 
	 * */
	
	size_t newSizeRequired = mBytesUsed + strlen(variableName) + 1 + 8;
	
	if (newSizeRequired >= mAllocatedPacketSize - 2) {
		if (mAllocatedPacketSize >= kTelemPacketMaxSize) {
			semGive(mPacketSem);
			return; // we will not store the variable if we've reached the max packet size
		}
		
		// Determine the new packet size
		if (mAllocatedPacketSize + kTelemPacketExpansionIncrement > kTelemPacketMaxSize)
			mAllocatedPacketSize = kTelemPacketMaxSize;
		else
			mAllocatedPacketSize += kTelemPacketExpansionIncrement;

		mVarsStart = (char *) realloc(mVarsStart, mAllocatedPacketSize);
	}
	
	if (newSizeRequired < mAllocatedPacketSize) {
		// we allocated more space, but still not enough to store this variable. Probably the variable name is too long. 
		
		semGive(mPacketSem);
		return;
	}
	
	char *nextVariableStart = mVarsStart + mBytesUsed;
	
	/*** Copy the name ***/
	size_t lengthToCopy = strlen(variableName); // we do not copy the terminating '\0': the protocol requires us to use a colon (:) instead
	
	char *colonLoc = strchr(variableName, ':');
	if (colonLoc != NULL) {
		// found a colon
		lengthToCopy = colonLoc - variableName; // make sure we don't copy any colons in the variable name as this would confuse PQTelemetry
	}
	
	strncpy(nextVariableStart, variableName, lengthToCopy);
	
	*(nextVariableStart + lengthToCopy) = ':'; // add the colon
	
	/*** Copy the double ***/
	memcpy(nextVariableStart + lengthToCopy + 1, &value, sizeof(double));
	
	mBytesUsed += lengthToCopy + 1 + 8; // variable uses however many bytes the name is plus 1 for the colon plus 8 for the double
	
	semGive(mPacketSem);
}

void Telemetry::sendPacket() {

#ifndef ENABLE_TELEMETRY
	return;
#endif

	// Take the semaphore so we can modify the packet pointer
	semTake(mPacketSem, WAIT_FOREVER);
	char *currentPacket = mVarsStart;
	size_t packetSize = mBytesUsed;

	startNewPacket();

	/* Now that the new packet has been created,
	 * give back the semaphore so addVariable can run. This ensures the
	 * time-consuming step of sending the packet doesn't block addVariable calls. */
	semGive(mPacketSem);

	if (sendto(sFd, currentPacket, packetSize, 0,
			(struct sockaddr *) &serverAddr, sockAddrSize) == ERROR) {
		// printf("Error sending packet\n");
		/*switch (errno) {
		case EISCONN:
			printf("EISCONN\n");
			break;
		case EMSGSIZE:
			printf("EMSGSIZE: %i\n", packetSize);
			break;
		case ENOBUFS:
			printf("ENOBUFS\n");
			break;
		case ENOMEM:
			printf("ENOMEM\n");
			break;
		case EBADF:
			printf("EBADF\n");
			break;
		case EFAULT:
			printf("EFAULT\n");
			break;
			
		 default:
		 printf("Other error: %i\n", errno);
		 
		}*/
	}

	free(currentPacket);
}
