/*
 * PQTelemetry v5.0 for VxWorks, Team 1717
 * This class transmits your data to PQTelemetry,
 * 	a graphing program that can be run on your laptop. 
 * 	You MUST use PQTelemetry >= v5 with this verion of Telemetry!
 * 	
 * 	Note that this program was written for the cRIO, which is PPC and therefore
 * 	big-endian. PenguinGrapher v3 must reverse the byte order since it runs on
 * 	Intel processors, which are little-endian. 
 * 	
 * 	This software was designed for use with FRC's WPILib but it could easily
 * 	be used with any VxWorks. You would only need to modify the code to give it
 * 	another source of time than a WPILib Timer (so that PQTelemetry's x-axis
 * 	can be correctly calculated.)
 * 	
 * 	
 *  WARNING: Do not send a variable named "time"--this identifier is used
 *  internally by PQTelemetry. 
 *  
 *  
 * HISTORY
 * ==============================
 * Originally written by Kevin Wojcik and Jake Moghtader, 2011?. 
 * Rewritten from scratch by Daniel Richman, 2013. 
 * 
 * Telemetry v5:
 *  - New protocol means further space saved in data transmission. 
 * 
 * Telemetry v3:
 * 	- Significantly faster than original due to better use of memory
 * 	- Thread-safe and fast for tuning PIDControllers
 * 	
 * 	
 * LICENSE
 * ===============================
 * Copyright (c) June 2013, Daniel Richman
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * To facilitate integrating this software into your own projects, redistributions in binary form need not retain any of this
 * 		license text, but the below disclaimer still applies. Attribution would be appreciated nonetheless!
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef TELEMETRY_H_
#define TELEMETRY_H_

#define ENABLE_TELEMETRY

#include <inetLib.h> // struct serverAddr
#include <semLib.h> // for semaphores
#include "Timer.h"

class Telemetry {
public:
	static Telemetry* getSingletonPtr();
	
	// init: call this once in RobotInit or similar
	// You can pass in your own timer; if NULL, Telemetry will create its own
	void init(char *serverIP, Timer *timer = NULL);
	
	void addVariable(char *variableName, double value); // thread-safe
	void sendPacket(); // do NOT call in more than one thread (but safe to use with addVariable in multiple threads)
	
private:
	void startNewPacket(); // this method is now internal to Telem; no need to call it externally
	
	static Telemetry* mInstance;
	
	// socket stuff
	struct sockaddr_in  serverAddr;    /* server's socket address */ 
	int                 sockAddrSize;  /* size of socket address structure */ 
	int                 sFd;           /* socket file descriptor */  
	
	// hold the variables
	//mVars: all values are stored in here, then the entire block of memory is sent as binary
	char		 		*mVarsStart;
	size_t				mBytesUsed;
	size_t				mAllocatedPacketSize; // size of memory currently allocated for mVars
	
	// timer
	Timer 				*mTimer;
	
	// semaphores
	SEM_ID mPacketSem; // ensures the packet is only being modified by one caller at a time
};

#endif
